#!/usr/bin/env python

# Copyright 2012-2015 Intel Corporation All Rights Reserved.
# 
# The source code contained or described herein and all documents
# related to the source code ("Material") are owned by Intel
# Corporation or its suppliers or licensors. Title to the Material
# remains with Intel Corporation or its suppliers and licensors. The
# Material may contain trade secrets and proprietary and confidential
# information of Intel Corporation and its suppliers and licensors,
# and is protected by worldwide copyright and trade secret laws and
# treaty provisions. No part of the Material may be used, copied,
# reproduced, modified, published, uploaded, posted, transmitted,
# distributed, or disclosed in any way without Intel's prior express
# written permission.
# 
# No license under any patent, copyright, trade secret or other
# intellectual property right is granted to or conferred upon you by
# disclosure or delivery of the Materials, either expressly, by
# implication, inducement, estoppel or otherwise. Any license under
# such intellectual property rights must be express and approved by
# Intel in writing.
# 
# Unless otherwise agreed by Intel in writing, you may not remove or
# alter this notice or any other notice embedded in Materials by Intel
# or Intel's suppliers or licensors in any way.

from optparse import OptionParser
from xml.dom import minidom
import sys, os, time, subprocess, signal

class shell_colors:
  HEADER = '\033[95m'
  OKBLUE = '\033[94m'
  OKGREEN = '\033[92m'
  WARNING = '\033[93m'
  FAIL = '\033[91m'
  ENDC = '\033[0;0m'
  BOLD = '\033[1m'
  UNDERLINE = '\033[4m'

current_process = None

ANALYZER = os.path.dirname(os.path.realpath(__file__)) + os.sep + "intel64" + \
    os.sep + "clck-analyze"
COLLECT = os.path.dirname(os.path.realpath(__file__)) + os.sep + "intel64" + \
    os.sep + "clck-collect"
ANALYZER_XML = os.path.dirname(os.path.realpath(__file__)) + os.sep + ".." + \
    os.sep + "etc" + os.sep + "clck.xml"
COLLECT_XML = os.path.dirname(os.path.realpath(__file__)) + os.sep + ".." + \
    os.sep + "etc" + os.sep + "clckd.xml"

mode_map = ['', 'health', 'certification', '', 'compliance']
verbosity_map = ['', 'critical', 'error', 'warning', 'notice', 'info']

TMP_DIR = r'/tmp'

def kill_children(pid, pid_list):
  pid = str(pid)
  proc = subprocess.Popen('pgrep -d, -P '+pid, shell=True,
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  (out, err) = [x.strip() for x in proc.communicate()]

  if len(out) == 0:
    return

  child_pids = [x.strip() for x in out.strip().split(',')]

  for cpid in child_pids:
    pid_list.append(cpid)
    kill_children(cpid, pid_list)

    os.kill(int(cpid), signal.SIGTERM)

def handler(signum, frame):
  global current_process

  pid_list = [current_process.pid]
  kill_children(current_process.pid, pid_list)
  os.kill(current_process.pid, signal.SIGTERM)

  sys.exit()

signal.signal(signal.SIGINT, handler)

def set_mode(xmldoc, mode):
  doc_root = xmldoc.firstChild

  config_found = False
  clck_mode_found = False

  for doc_child in doc_root.childNodes:
    if doc_child.nodeName == "config":
      config_found = True
      config_root = doc_child

      for config in config_root.childNodes:
        if config.nodeName == "clck-mode":
          clck_mode_found = True
          config.firstChild.data = mode
          break

      if clck_mode_found == False:
        newElem = xmldoc.createElement("clck-mode")
        newElem.appendChild(xmldoc.createTextNode(mode))
        config_root.appendChild(newElem)


  if config_found == False:
    newInnerElem = xmldoc.createElement("clck-mode")
    newInnerElem.appendChild(xmldoc.createTextNode(mode))
    newElem = xmldoc.createElement("config")
    newElem.appendChild(newInnerElem)
    doc_root.appendChild(newElem)

  return xmldoc

def set_verbosity(xmldoc, level):
  doc_root = xmldoc.firstChild

  for doc_child in doc_root.childNodes:
    if doc_child.nodeName == "log_level":
      doc_child.firstChild.data = level
      break
  else:
    newLevel = xmldoc.createElement("log_level")
    newLevel.appendChild(xmldoc.createTextNode(level))
    doc_root.appendChild(newLevel)

  return xmldoc

def get_mod_list(xmldoc):
  doc_root = xmldoc.firstChild
  list = []

  for doc_child in doc_root.childNodes:
    if doc_child.nodeName == "config":
      config_root = doc_child

      for slot_name in config_root.childNodes:
        if slot_name.nodeName == "clck-checks":
          checks_root = slot_name

          for entry in checks_root.childNodes:
            if entry.nodeName == "entry":
              list.append(entry.firstChild.data.strip())


  return list

def del_mods(xmldoc, mod_names):
  doc_root = xmldoc.firstChild

  for doc_child in doc_root.childNodes:
    if doc_child.nodeName == "config":
      config_root = doc_child

      for slot_name in config_root.childNodes:
        if slot_name.nodeName == "clck-checks":
          checks_root = slot_name

          for entry in checks_root.childNodes:
            if entry.nodeName == "entry" and \
                entry.firstChild.data.strip() in mod_names:
                  checks_root.removeChild(entry)

  return xmldoc

def add_mods(xmldoc, mod_names):
  doc_root = xmldoc.firstChild

  for mod_name in mod_names:
    if mod_name in get_mod_list(xmldoc): continue

    newEntry = xmldoc.createElement("entry")
    newEntry.appendChild(xmldoc.createTextNode(mod_name))

    newClckCheck = xmldoc.createElement("clck-checks")
    newConfig = xmldoc.createElement("config")


    config_tag = None
    clck_checks_tag = None

    for doc_child in doc_root.childNodes:
      if doc_child.nodeName == "config":
        config_tag = doc_child

        for slot_name in config_tag.childNodes:
          if slot_name.nodeName == "clck-checks":
            clck_checks_tag = slot_name

            slot_name.appendChild(newEntry)
            break


  if config_tag == None:
    newConfig = xmldoc.createElement("config")
    newClckCheck = xmldoc.createElement("clck-checks")
    newClckCheck.appendChild(newEntry)
    newConfig.appendChild(newClckCheck)
    doc_root.appendChild(newConfig)
  elif clck_checks_tag == None:
    newClckCheck = xmldoc.createElement("clck-checks")
    newClckCheck.appendChild(newEntry)
    config_tag.appendChild(newClckCheck)

  return xmldoc

def create_tmp_file(tmp_dir):
  try:
    os.makedirs(tmp_dir)
  except OSError, e:
    if e.errno == 13:
      return None

  filename  = tmp_dir + os.sep + "clck-" + time.strftime("%Y%m%d-%H%M%S")

  try:
    file = open(filename, 'w')
  except OSError, e:
    if e.errno == 13:
      return None
  else:
    file.close()
    os.remove(filename)

  return filename

def get_and_verify_options():
  epilog_string = """
Mandatory or optional arguments to long options are also mandatory or
optional  for any corresponding short options.

Report bugs to <cluster@intel.com>
"""

  usage_string = """
clck [OPTION...]
clck -- Intel(R) Cluster Checker
"""

  OptionParser.format_epilog = lambda self, formatter: self.epilog
  parser = OptionParser(add_help_option = False, epilog = epilog_string,
      usage = usage_string)

  parser.add_option("-c", "--config", dest = "xmlfile", metavar = "FILE",
      help = "config file to be used", action = "store")

  parser.add_option("-C", "--certification", dest = "certification",
      metavar = "VERSION", help = "Run as required to certify Intel(R) Cluster"
      + "Ready", action = "store")

  parser.add_option("-d", "--debug", dest = "debug", action = "store_true",
      help = "Display database information for each generated sign")

  parser.add_option("-D", "--nodb", dest = "nodb", action = "store_true",
      help = "(unsupported)")

  parser.add_option("-e", "--exclude", dest = "exclude", action = "append",
      help = "Exclude a connector extension from being executed", metavar = "MOD")

  parser.add_option("-f", "-F", "--nodelist", "--nodefile", dest = "nodefile",
          action = "store", help = "Machine file with list of nodes",
          metavar = "FILE")

  parser.add_option("-H", "--noheader", dest = "noheader", action = "store_true",
      help = "(unsupported)")

  parser.add_option("-i", "--include", dest = "include_also", action = "append",
      help = "Add a connector extension to be executed", metavar = "MOD")

  parser.add_option("-I", "--include_only", dest = "include_only", action = "append",
      help = "Only execute a specific connector extension ", metavar = "MOD")

  parser.add_option("-k", "--compliance", dest = "compliance", action = "store",
      help = "Check compliance of the cluster to the Intel(R) Cluster Ready"
      + " Specification indicated by version", metavar = "MOD")

  parser.add_option("-L", "--level", dest = "level", action = "store",
      type = "int", metavar = "[1-5]",
      help = "Levels 1 through 4 execute all test modules with the exception" + \
          " of High Performance Linpack. Level 5 executes the High" + \
          " Performance Linpack test module as well")

  parser.add_option("-n", "--nodeps", dest = "nodeps", action = "store_true",
      help = "(unsupported)")

  parser.add_option("-N", "--nolog", dest = "nolog", action = "store_true",
      help = "Do not write the output log to disk")

  parser.add_option("-t", "--health", dest = "health", action = "store_true",
      help = "Check cluster health")

  parser.add_option("-v", "--verbose", dest = "verbosity", action = "store",
      help = "Produce verbose output. Higher values produce more output",
      metavar = "[1-5]", type = "int")

  parser.add_option("-h", "-?", "--usage", "--help", action = "store_true",
      dest = "help", help = "Give this help list")

  parser.add_option("-V", "--version", dest = "version", action = "store_true",
          help = "Print program version")

  (options, arguments) = parser.parse_args()

  if options.nodb:
    print shell_colors.WARNING + \
        "The --nodb (-D) option is unsupported and does not have" + \
        " any functionality" + shell_colors.ENDC

  if options.certification != None and options.certification != "1.3.1":
    print(shell_colors.FAIL + \
        "Currently, only certification against Intel(R) Cluster Ready" + \
        " Specification version 1.3.1 is supported" + \
        shell_colors.ENDC)
    sys.exit()

  if options.compliance != None and options.compliance != "1.3.1":
    print(shell_colors.FAIL + \
        "Currently, only compliance against Intel(R) Cluster Ready" + \
        " Specification version 1.3.1 is supported" + \
        shell_colors.ENDC)
    sys.exit()

  mode = 0 
  if options.health != None: mode = mode + (1)
  if options.certification != None: mode = mode + (1 << 1)
  if options.compliance != None: mode = mode + (1 << 2)
  if mode == 0: mode = 1

  options.mode = mode

  if mode != 1 and mode != 2 and mode != 4:
    print (shell_colors.FAIL + \
        "Mode can either be --certification (-C), --compliance (-k) or " + \
        "health (-t). Setting more than one mode is not supported"
        + shell_colors.ENDC)
    sys.exit()

  if options.help:
    parser.print_help()
    sys.exit()

  if options.version:
    os.system(ANALYZER + " --version")
    sys.exit()

  if options.nodeps:
    print shell_colors.WARNING + \
        "The --nodeps (-n) option is unsupported and does not have" + \
        " any functionality" + shell_colors.ENDC

  if options.noheader:
    print shell_colors.WARNING + \
        "The --noheader (-H) option is unsupported and does not have" + \
        " any functionality" + shell_colors.ENDC

  if options.include_only != None  and options.exclude != None:
    print(shell_colors.FAIL + \
        "The options --include_only and --exclude can not both be used" + \
        " simultaneously" + \
        shell_colors.ENDC)
    sys.exit()

  if options.include_only == None: options.include_only = []
  if options.include_also == None: options.include_also = []
  if options.exclude == None: options.exclude = []

  if options.level != None and (options.level < 1 or options.level > 5):
    print(shell_colors.FAIL + \
        "Valid values for --level(-L) are between 1 and 5" + \
        shell_colors.ENDC)
    sys.exit()

  if options.level == None: options.level = -1  

  if options.verbosity != None and \
      (options.verbosity < 1 or options.verbosity > 5):
    print(shell_colors.FAIL + \
        "Valid values for --verbose(-V) are between 1 and 5" + \
        shell_colors.ENDC)
    sys.exit()

  if options.verbosity == None: options.verbosity = -1

  if options.nodefile == None:
    print (shell_colors.FAIL + \
        "A nodefile must be specified via --nodefile(-F)" + \
        shell_colors.ENDC)
    sys.exit()

  if options.nolog == None: options.nolog == -1
  return options

def update_xml_contents(options):
  filename = options.xmlfile

  if filename == None:
      filename = ANALYZER_XML

  xmldoc = None

  try:
    xmldoc = minidom.parse(filename)
  except:
    print(shell_colors.FAIL + \
        "Unable to parse XML file" + \
        shell_colors.ENDC)
    sys.exit()

  xmldoc = set_mode(xmldoc, mode_map[options.mode])

  if options.verbosity != -1:
    xmldoc = set_verbosity(xmldoc, verbosity_map[options.mode])

  # If the user chooses include_only, options could be include_only, include_also
  # If not, options could be exclude, include
  if len(options.include_only) > 0:
    mod_list = get_mod_list(xmldoc)
    xmldoc = del_mods(xmldoc, mod_list)

    xmldoc = add_mods(xmldoc, options.include_only)

    if len(options.include_also) > 0:
      xlmdoc = add_mods(xmldoc, options.include_also)

  else:
    if len(options.include_also) > 0:
      xmldoc = add_mods(xmldoc, options.include_also)
    if len(options.exclude) > 0:
      xmldoc = del_mods(xmldoc, options.exclude)

  return xmldoc

def run_collect(options):
  global current_process
  full_command = COLLECT

  full_command = full_command + " -f " + options.nodefile

  if options.verbosity != -1:
    full_command = full_command + " -l " + verbosity_map[options.verbosity]

  if options.xmlfile == None:
    full_command = full_command + " -c " + COLLECT_XML
  else:
    full_command = full_command + " -c " +  options.xmlfile

  if options.level == 5:
    full_command = full_command + " -a"
  else:
    full_command = full_command + " -s exclude_hpl"

  print "Starting data collection"
  current_process = subprocess.Popen(full_command, shell = True, preexec_fn = os.setsid)
  current_process.wait()

def run_analyze(options):
  full_command = ANALYZER

  full_command = full_command + " -f " + options.nodefile

  if options.nolog == None and options.output_filename != None:
    full_command = full_command + " -o " + options.output_filename

  full_command = full_command + " -c " + options.tmp_xml_filename

  current_process = subprocess.Popen(full_command, shell = True, preexec_fn = os.setsid)
  current_process.wait()

# -------------------------------------------------------------------------------

if __name__ == "__main__":

  options = get_and_verify_options()

  options.output_filename = None

  if options.nolog == -1 :

    log_dir = os.getenv('CLCK_LOG_DIR', TMP_DIR)
    output_filename = create_tmp_file(log_dir)

    if output_filename == None:
      print(shell_colors.FAIL + \
          "Unable to create output log file. Please check directory permissions" + \
          shell_colors.ENDC)
    else:
      options.output_filename = output_filename + ".out"

  run_collect(options)

  xmldoc = update_xml_contents(options)

  tmp_dir = os.getenv('CLCK_LOG_DIR', TMP_DIR)
  tmp_xml_filename = create_tmp_file(tmp_dir) + ".xml"

  with open(tmp_xml_filename, 'w') as f: f.write(xmldoc.toxml())

  options.tmp_xml_filename = tmp_xml_filename

  run_analyze(options)

  os.remove(tmp_xml_filename)

  if options.output_filename != None:
    print shell_colors.OKBLUE + \
        "Log file saved in " + options.output_filename + \
        shell_colors.ENDC
