/*
 * Copyright 2012-2015 Intel Corporation All Rights Reserved.
 *
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel
 * Corporation or its suppliers or licensors. Title to the Material
 * remains with Intel Corporation or its suppliers and licensors. The
 * Material may contain trade secrets and proprietary and confidential
 * information of Intel Corporation and its suppliers and licensors,
 * and is protected by worldwide copyright and trade secret laws and
 * treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way without Intel's prior express
 * written permission.
 *
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise. Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * Unless otherwise agreed by Intel in writing, you may not remove or
 * alter this notice or any other notice embedded in Materials by Intel
 * or Intel's suppliers or licensors in any way.
 */

#ifndef CLCK_H
#define CLCK_H

#include <bitset>
#include <condition_variable>
#include <memory>
#include <set>
#include <string>
#include <vector>

namespace clck {
#define CLCK_LAYER_VERSION_NUMBER 3000000

  /**
   * @brief Base class for database configuration
   */
  class Database {
  public:
    virtual ~Database() = 0; /**< Database destructor */
  };

  /**
   * @brief SQLite configuration.  Derived from Database.
   */
  class SQLite : public Database {
  public:
    SQLite(const std::string &db_file); /**< SQLite constructor */
    ~SQLite();                          /**< SQLite destructor */

    std::string db_file; /**< Absolute path to the database file */
  };

  /**
   * @brief A fault is the basic analysis unit.  A fault is either a
   *        sign (i.e., observation) or a diagnosis (i.e., root
   *        cause).
   */
  class Fault {
  public:
    int confidence = 0;             /**< Confidence percentage (0 - 100) */
    std::string id;                 /**< Message catalog id */
    std::string msg;                /**< Expanded message string */
    std::vector<std::string> nodes; /**< List of nodes */
    std::string remedy;             /**< Expanded remedy string */
    int severity = 0;               /**< Severity percentage (0 - 100) */
    bool suppressed = false;        /**< True if the diagnosis / sign is
                                         suppressed, false otherwise */
    std::set<int> rowid;            /**< DB rows that provide the raw data
                                         upon which the sign/diagnosis is based */
  };

  /**
   * @brief A diagnosis is the root cause of an issue.  Diagnosis is
   *        derived from the Fault class.
   *
   * @todo A new type field will be added to the Fault class.
   *       This class will be removed.
   */
  class Diagnosis : public Fault {
  public:

  };

  /**
   * @brief A sign is an observation of an issue.  Sign is derived
   *        from the Fault class.
   *
   * @todo A new type field will be added to the Fault class.
   *       This class will be removed.
   */
  class Sign : public Fault {
  public:
    enum { DIAGNOSED, OBSERVED }
      /**
       * Either diagnosed (meaning used to make a diagnosis) or
       * observed (undiagnosed).
       */
      state;
  };

  /**
   * @brief The node container.
   */
  class Node {
  public:
    /**
     * Roles than a node can fulfill.
     */
    enum role_t { ROLE_BOOT, ROLE_COMPUTE, ROLE_ENHANCED,
                  ROLE_EXTERNAL, ROLE_HEAD, ROLE_JOB_SCHEDULE, ROLE_LOGIN,
                  ROLE_NETWORK_ADDRESS, ROLE_STORAGE };

    /**
     * A subcluster to which the node belongs.
     */
    std::string subcluster;

    /**
     * An unique node identifier, i.e., the hostname.
     */
    std::string name;

    /**
     * A list of roles that the node fulfills.
     */
    std::vector<role_t> roles;
  };

  /**
   * @brief The presentation layer.
   */
  class Layer {
  public:
    /**
     * @brief Data analysis configuration parameter.
     */
    struct ConfigParam {
      std::string module;              /**< Module name in the knowledge base */
      std::string key;                 /**< Parameter name */
      std::vector<std::string> values; /**< Parameter value(s) */
    };

    /**
     * @brief Layer configuration options
     *
     * @todo Additional configuration options are likely to be added
     *       in the future, as needed.
     */
    struct Config {
      /**
       * Instance of class derived from Database
       */
      std::shared_ptr<Database> db;

      /**
       * Maximum allowable age of data, relative to the value of now.  Data
       * older than now minus expiration will be ignored.  A value of 0 means
       * to use all data, i.e., there is no expiration.
       */
      time_t expiration = 0;

      /**
       * List of connector extensions to be loaded
       */
      std::vector<std::string> extension_mods;

      /**
       * Absolute path to the connector extension directory
       */
      std::string extension_path;

      /**
       * List of knowledge base files to be loaded
       */
      std::vector<std::string> kb_mods;

      /**
       * Absolute path to the knowledge base directory
       */
      std::string kb_path;

      /**
       * String representing the message catalog language
       */
      std::string language;

      /**
       * List of nodes to be analyzed (see @ref node_source).
       */
      std::vector<Node> nodes;

      enum { DATABASE, NODELIST, INTERSECTION, UNION }
        /**
         * Select the source for the list of nodes.  Options are to use
         * the database (exclusively), the input @ref nodes vector
         * (exclusively), or the intersection or union of the two.
         */
        node_source = DATABASE;

      /**
       * Reference time to be used as the current time for analysis
       */
      time_t now = time(NULL);

      /**
       * Data analysis configuration parameters
       */
      std::vector<ConfigParam> config_params;

      /**
       * Default Config constructor
       */
      Config() = default;

      /**
       * Config constructor
       *
       * @param db             Instance of class derived from Database
       * @param extension_mods List of connector extensions to be loaded
       * @param kb_mods        List of knowledge base files to be loaded
       * @param extension_path Absolute path to the connector extension directory
       * @param kb_path        Absolute path to the knowledge base directory
       */
      Config(std::shared_ptr<Database> db,
             const std::vector<std::string> &extension_mods,
             const std::vector<std::string> &kb_mods,
             const std::string &extension_path = "",
             const std::string &kb_path = "");
    };

    /**
     * @brief Filter for the list of faults returned by get_faults().
     */
    struct Filter {
      /**
       * Select faults with a greater than or equal to confidence value.
       */
      int confidence = 0;

      /**
       * Select faults corresponding to at least one of the ids.  If empty,
       * does not filter on id.
       */
      std::vector<std::string> ids;

      /**
       * Selects faults corresponding to at least one of the nodes.  If empty,
       * does not filter on node.
       */
      std::vector<std::string> nodes;

      /**
       * Select faults with a greater than or equal to severity value.
       */
      int severity = 0;

#define CLCK_FAULT_STATE_DIAGNOSED  1
#define CLCK_FAULT_STATE_OBSERVED   2
      /**
       * Select faults with a matching state.  Only applies to signs,
       * not diagnoses.
       */
      std::bitset<2> state = CLCK_FAULT_STATE_DIAGNOSED | CLCK_FAULT_STATE_OBSERVED;

#define CLCK_FAULT_SUPPRESSED_TRUE  1
#define CLCK_FAULT_SUPPRESSED_FALSE 2
      /**
       * Select faults with a matching suppression value.
       */
      std::bitset<2> suppressed = CLCK_FAULT_SUPPRESSED_FALSE;

#define CLCK_FAULT_TYPE_DIAGNOSIS 1
#define CLCK_FAULT_TYPE_SIGN      2
      /**
       * Select faults with a matching type.
       */
      std::bitset<2> type = CLCK_FAULT_TYPE_DIAGNOSIS | CLCK_FAULT_TYPE_SIGN;
    };

    /**
     * @brief Internal Layer messages for the caller to handle
     */
    struct Message {
      int level;       /**< Message level (priority).
                            Inherits logging levels from syslog. */
      std::string msg; /**< Message string */
    };

    /**
     * @brief Sort order for the list of faults returned by
     *        get_faults().
     */
    struct Sorting {
      /**
       * If true, sort in ascending order, otherwise sort in descending order.
       */
      bool ascending = true;

      enum { CONFIDENCE, ID, NODE, SEVERITY }
        field; /**< Fault field to sort on */

      /**
       * Sorting constructor
       */
      Sorting(bool ascending, decltype(field) field);
    };

    /**
     * @brief Suppress faults matching the specified values.
     */
    struct Suppression {
      /**
       * Suppress all faults with a value less than the confidence value.
       */
      int confidence = 0;

      /**
       * Suppress all messages with a matching id.  If empty, does not
       * suppress on id.
       */
      std::string id;

      /**
       * Supress all messages containing the node value.  If empty, does not
       * suppress on node.
       */
      std::string node;

      /**
       * Suppress all faults with a value less than the severity value.
       */
      int severity = 0;
    };

    /** Notify when a new internal Layer message is available */
    std::condition_variable message;

    /**
     * Layer constructor
     */
    Layer(const Config &config);

    /**
     * Layer destructor
     */
    ~Layer();

    /**
     * Start the analysis.  Note: behavior is undefined if invoked more than
     * once per Layer instance.
     *
     * @return True if the analysis was completed successfully, false
     *         otherwise.  Note: this does NOT reflect the state of the
     *         cluster.
     */
    bool analyze(const std::vector<Suppression> &suppressions);

    /**
     * Collect data.
     *
     * @todo Placeholder - not implemented
     *
     * @return True if the data collection was completed successfully,
     *         false otherwise.
     */
    bool collect() { return false; }

    /**
     * Returns a list of faults.  May be called multiple times.  While analyze()
     * is active, presumably in another thread, the behavior is undefined.
     * Use progress() to determine current analysis status.  If called before
     * analyze(), returns an empty list.
     *
     * @param filter Return only faults that match the filter
     *
     * @param sorting List of sorting criteria.  The first element is the
     *          primary sorting criterion, the second element the
     *          secondary sorting criterion, etc.  If empty, the
     *          faults are returned unsorted.
     *
     * @return A list of faults
     */
    std::vector<std::shared_ptr<Fault>>
    get_faults(const Filter &filter, const std::vector<Sorting> &sorting);

    /**
     * Returns a list of messages generated internal to Layer for the caller to
     * handle.
     *
     * @return A list of messages
     */
    std::vector<Message> get_messages();

    /**
     * Returns the list of nodes to be analyzed.
     *
     * @return A list of nodes.
     */
    std::vector<Node> get_nodes();

    /**
     * Returns the version of the Layer API.
     *
     * @return The version number.  Version X.Y.Z is represented as
     *         (X*1000000 + Y*1000 + Z).
     */
    int get_version_number();

    /**
     * While analyze() is active, presumably in another thread, returns
     * the number of rules remaining to be fired and the number of rules already
     * run.  If called before analyze(), both values will be 0.
     *
     * @param remaining The number of rules remaining to be fired.
     *                  Not guaranteed to be monotonic.  Returned by
     *                  reference.
     *
     * @param completed The number of rules that have been fired.
     *                  Will be monotonic.  Returned by reference.
     *
     * @return False if analyze() has not yet been called, true
     *         otherwise.
     */
    bool progress(unsigned long &remaining, unsigned long &completed);

    class Impl;

  private:
    std::unique_ptr<Impl> pImpl;
  };
}

#endif // CLCK_H
