 ; Copyright 2012-2015 Intel Corporation All Rights Reserved.
 ;
 ; The source code contained or described herein and all documents
 ; related to the source code ("Material") are owned by Intel
 ; Corporation or its suppliers or licensors. Title to the Material
 ; remains with Intel Corporation or its suppliers and licensors. The
 ; Material may contain trade secrets and proprietary and confidential
 ; information of Intel Corporation and its suppliers and licensors,
 ; and is protected by worldwide copyright and trade secret laws and
 ; treaty provisions. No part of the Material may be used, copied,
 ; reproduced, modified, published, uploaded, posted, transmitted,
 ; distributed, or disclosed in any way without Intel's prior express
 ; written permission.
 ;
 ; No license under any patent, copyright, trade secret or other
 ; intellectual property right is granted to or conferred upon you by
 ; disclosure or delivery of the Materials, either expressly, by
 ; implication, inducement, estoppel or otherwise. Any license under
 ; such intellectual property rights must be express and approved by
 ; Intel in writing.
 ;
 ; Unless otherwise agreed by Intel in writing, you may not remove or
 ; alter this notice or any other notice embedded in Materials by Intel
 ; or Intel's suppliers or licensors in any way.
(defclass _MULTISET
  "Class to represent the multiset (aka bag) data structure"
  (is-a USER)
  (role concrete)

  (multislot element (visibility private))
  (multislot multiplicity (type INTEGER) (visibility private)))

(defmessage-handler _MULTISET add (?v)
  "Add a new data element.  If the element does not already exist,
   create it."
  (bind ?pos (member$ ?v ?self:element))
  (if ?pos then
    ; element already exists, just increment the multiplicity
    (slot-replace$ ?self multiplicity ?pos ?pos
                   (+ 1 (nth$ ?pos ?self:multiplicity)))
    else
    ; element needs to be added
    (slot-insert$ ?self element 1 ?v)
    (slot-insert$ ?self multiplicity 1 1)))

(defmessage-handler _MULTISET cardinality ()
  "The cardinality is the sum of all the multiplicities."
  (bind ?sum 0)
  (progn$ (?i ?self:multiplicity)
          (bind ?sum (+ ?i ?sum)))
  (return ?sum))

(defmessage-handler _MULTISET fraction (?v)
  "Compute the multiplicity fraction corresponding to an element.
   This is equivalent to the Jaccard index where A is the original
   multiset and B is the multiset containing only the specified
   element.

   The return value is bounded between 0 and 1.  If the multiset is
   uniform, then this returns 1.  If the element does not appear in
   the multiset, then this returns 0.

   For example:
     A = {a:8, b:1, c:1}
     B = {a:8}

     J(A,B) = |intersection(A,B)| / |union(A,B)|

     |intersection({a:8, b:1, c:1}, {a:8})| = |{a:8}| = 8
     |union({a:8, b:1, c:1}, {a:8})| = |{a:8, b:1, c:1}| = 10

     J(A,B) = 8 / 10 = 0.8"
  (return (/ (send ?self multiplicity ?v) (send ?self cardinality))))

(defmessage-handler _MULTISET multiplicity (?v)
  "Get the multiplicity corresponding to an element."
  (bind ?pos (member ?v ?self:element))
  (if ?pos
      then (return (nth$ ?pos ?self:multiplicity))
      else (return 0)))

(defclass MULTISET
  "Inheriting this class provides a class with access to multiset
   capabilities.  The slots in the class are shared so the slots are
   globally accessible from all inheriting classes.

   The class is a key / value map.  Keys may be arbitrarily defined
   and map to instances of the _MULTISET class.  Any number of keys
   may be created per class, but keys are global, so it's highly
   recommended that the key include the name of the class.  If per
   node multisets are needed, the key should also include the node_id.
   Further, if multisets are needed for multiple slots in the same
   class, include the slot name in the key.  For example, the PYTHON
   class keys might be PYTHON-node00-version and PYTHON-node00-stderr.

   One-to-one wrapper message handlers are provided for the _MULTISET
   message handlers.

   To gain multiset capabilities, a class should do the following:

     - Inherit MULTISET
       E.g.,
       (defclass FOO (is-a MYCLASS ... MULTISET)
         ...
         (slot bar)
         ...

     - Wrap the put message handler for each of the class slots
       wanting multiset capabilities to the value using one or more
       keys.
       E.g.,

       (defmessage-handler FOO put-bar after (?v)
         (bind ?key (class ?self))
         (send ?self add ?key ?v))

       Alternatively, the init message handler could be wrapped:

       (defmessage-handler FOO init after ()
         (bind ?key (class ?self))
         (send ?self add ?ke ?self:bar))

   Note: modifying or removing keys once they have been added is not
   supported."
  (is-a USER)
  (role abstract)

  (slot multiset_control (default FALSE) (storage shared))
  (multislot _multiset_keys (type SYMBOL)
             (pattern-match non-reactive) ; critical for performance
             (storage shared)
             (visibility private))
  (multislot _multiset_instances (type INSTANCE)
             (allowed-classes _MULTISET)
             (pattern-match non-reactive) ; critical for performance
             (storage shared)
             (visibility private)))

(defrule flip_multiset_control
  "Rules start matching immediately even though additional data points
   may have yet to be created.  To force rules to delay until all the
   instances have been defined, use a control variable.  This rule
   will always activate due to the empty left hand side, but the
   control variable will not be changed to the TRUE state until the
   run command is issued.  Since the slot is shared, flipping the
   control variable in any one instance will do."
  =>
  (do-for-instance ((?i MULTISET)) TRUE
                   (send ?i put-multiset_control TRUE)))

(defmessage-handler MULTISET _find_pair (?key)
  "Helper to look up the _MULTISET instance corresponding to the
   key."
  (bind ?pos (member$ ?key ?self:_multiset_keys))
  (if ?pos
      then (return (nth$ ?pos ?self:_multiset_instances))
      else (return FALSE)))

(defmessage-handler MULTISET add (?key ?v)
  "Add the value to corresponding _MULTISET instance.  If the key /
   value pair does not already exist, create it."
  (bind ?pos (member$ ?key ?self:_multiset_keys))
  (if ?pos then
    ; key/instance pair already exists
    (bind ?instance (nth$ ?pos ?self:_multiset_instances))
    else
    ; key/instance pair needs to be created and added
    (bind ?instance (make-instance of _MULTISET))
    (slot-insert$ ?self _multiset_keys 1 ?key)
    (slot-insert$ ?self _multiset_instances 1 ?instance))
  (send ?instance add ?v))

(defmessage-handler MULTISET cardinality (?key)
  "Wrap the corresponding _MULTISET message handler"
  (bind ?instance (send ?self _find_pair ?key))
  (if ?instance then (send ?instance cardinality)))

(defmessage-handler MULTISET multiplicity (?key ?v)
  "Wrap the corresponding _MULTISET message handler"
  (bind ?instance (send ?self _find_pair ?key))
  (if ?instance then (send ?instance multiplicity ?v)))

(defmessage-handler MULTISET fraction (?key ?v)
  "Wrap the corresponding _MULTISET message handler"
  (bind ?instance (send ?self _find_pair ?key))
  (if ?instance then (send ?instance fraction ?v)))
