; Copyright 2006-2015 Intel Corporation All Rights Reserved.
;
; The source code contained or described herein and all documents
; related to the source code ("Material") are owned by Intel
; Corporation or its suppliers or licensors. Title to the Material
; remains with Intel Corporation or its suppliers and licensors. The
; Material may contain trade secrets and proprietary and confidential
; information of Intel Corporation and its suppliers and licensors,
; and is protected by worldwide copyright and trade secret laws and
; treaty provisions. No part of the Material may be used, copied,
; reproduced, modified, published, uploaded, posted, transmitted,
; distributed, or disclosed in any way without Intel's prior express
; written permission.
;
; No license under any patent, copyright, trade secret or other
; intellectual property right is granted to or conferred upon you by
; disclosure or delivery of the Materials, either expressly, by
; implication, inducement, estoppel or otherwise. Any license under
; such intellectual property rights must be express and approved by
; Intel in writing.
;
; Unless otherwise agreed by Intel in writing, you may not remove or
; alter this notice or any other notice embedded in Materials by Intel
; or Intel's suppliers or licensors in any way.

(defrule dgemm-data-missing
  "Detect cases where there is no DGEMM data."
  (object (is-a CONFIG) (name [config]) (clck-mode certification | health)
          (clck-checks $? dgemm $?))
  (object (is-a NODE) (node_id ?node)
          (role $?role&:(member$ compute ?role)|:(member$ enhanced ?role)))
  (not (object (is-a DGEMM) (node_id ?node)))
  =>
  (make-instance of BOOLEAN_SIGN (node_id ?node)
                 (confidence 90) (severity 90)
                 (state observed) (value TRUE)
                 (id "dgemm-data-missing")
                 (remedy "dgemm-data-missing-remedy")))

(defrule dgemm-outlier
  "Locate values that are outliers.  An outlier is a value that is
   outside the range defined by the median +/- 6 * median absolute
   deviation.  The statistics are computed using all samples on all
   nodes (i.e., use the DGEMM statistics key).

   Note: the statistics_control condition is required to ensure that
   all samples are included when computing the statistics."
  (object (is-a CONFIG) (name [config]) (clck-mode certification | health)
          (clck-checks $? dgemm $?) (dgemm-number-of-mad ?num-mad))
  (object (is-a NODE) (node_id ?node)
          (role $?role&:(member$ compute ?role)|:(member$ enhanced ?role))
          (subcluster ?subcluster))
  ?obj <- (object (is-a DGEMM) (node_id ?node) (flops ?flops&~0)
                  (statistics_control TRUE))
  (test (or
         (> ?flops (+ (send ?obj median (send ?obj statistics-key))
                      (* ?num-mad (send ?obj mad (send ?obj statistics-key)))))
         (< ?flops (- (send ?obj median (send ?obj statistics-key))
                      (* ?num-mad (send ?obj mad (send ?obj statistics-key)))))))
  =>
  (bind ?key (send ?obj statistics-key))
  (bind ?median (send ?obj median ?key))
  (bind ?mad (* ?num-mad (send ?obj mad ?key)))
  (bind ?severity (* (- 1 (gaussian ?flops ?median
                                    (* 3 1.4826 (send ?obj mad ?key)))) 80))
  (make-instance of BOOLEAN_SIGN (node_id ?node)
                 (confidence 75) (severity ?severity) (state observed)
                 (id "dgemm-outlier") (source ?obj)
                 (args (create$ (/ ?flops 1e9) (/ ?median 1e9)
                                (/ ?mad 1e9) ?num-mad) ?subcluster)))

(defrule latest-dgemm-data-is-too-old
  "Identify nodes where the most recent DGEMM data should be
   considered too old.  Do not fire if there is not a recent
   heartbeat, i.e., the node is not returning any data and there is a
   not a DGEMM specific issue.  Too old is defined to mean no data
   from the last 7 days (604800 seconds)."
  (object (is-a CONFIG) (name [config]) (clck-mode certification | health)
          (clck-checks $? dgemm $?) (data-age-threshold ?age-threshold))
  (object (is-a NODE) (node_id ?node)
          (role $?role&:(member$ compute ?role)|:(member$ enhanced ?role)))
  ?obj <- (object (is-a DGEMM) (node_id ?node) (timestamp ?t))
  (not (object (is-a SIGN) (node_id ?node) (id "heartbeat-is-too-old")))
  (test (< ?t (- (now) ?age-threshold)))
  =>
  (bind ?age (round (- (now) ?t)))
  (make-instance of BOOLEAN_SIGN (node_id ?node) (confidence 90) (severity 20)
                 (state observed) (value TRUE) (source ?obj)
                 (id "latest-dgemm-data-is-too-old")
                 (args ?age)))

(defrule latest-dgemm-data-is-substandard
  "For the most recent DGEMM data point, identify nodes with
   substandard FLOPS relative to a threshold based on the hardware.
   The severity depends on the amount of deviation from the threshold
   value; the larger the deviation, the higher the severity."
  (object (is-a CONFIG) (name [config]) (clck-mode certification | health)
          (clck-checks $? dgemm $?)
          (dgemm-peak-fraction ?fraction))
  (object (is-a NODE) (node_id ?node)
          (role $?role&:(member$ compute ?role)|:(member$ enhanced ?role)))
  ?obj <- (object (is-a DGEMM) (flops ?flops&~0) (node_id ?node) (peak ?peak&~0))
  (test (< (/ ?flops ?peak) ?fraction))
  =>
  (bind ?severity (* (- 1 (gaussian ?flops (* ?fraction ?peak)
                                    (/ ?peak 10))) 80))
  (make-instance of PERFORMANCE_SIGN (node_id ?node) (confidence 90)
                 (severity ?severity) (state observed) (value substandard)
                 (id "latest-dgemm-data-is-substandard")
                 (source ?obj)
                 (args (create$ ; convert FLOPS to GFLOPS
                        (/ ?flops 1e9) (* 100 (/ ?flops ?peak))
                        (/ ?peak 1e9)))))

(defrule invalid-dgemm-data
  "Detect cases where the DGEMM data is invalid, i.e., data provider
   output exists in the database, but the connector could not parse
   it."
  (object (is-a CONFIG) (name [config]) (clck-mode certification | health)
          (clck-checks $? dgemm $?))
  (object (is-a NODE) (node_id ?node)
          (role $?role&:(member$ compute ?role)|:(member$ enhanced ?role)))
  ?obj <- (object (is-a DGEMM) (node_id ?node) (flops 0))
  =>
  (make-instance of BOOLEAN_SIGN (node_id (send ?obj get-node_id))
                 (confidence 60) (severity 90) (state observed) (source ?obj)
                 (value TRUE) (id "invalid-dgemm-data")))
