; Copyright 2006-2015 Intel Corporation All Rights Reserved.
;
; The source code contained or described herein and all documents
; related to the source code ("Material") are owned by Intel
; Corporation or its suppliers or licensors. Title to the Material
; remains with Intel Corporation or its suppliers and licensors. The
; Material may contain trade secrets and proprietary and confidential
; information of Intel Corporation and its suppliers and licensors,
; and is protected by worldwide copyright and trade secret laws and
; treaty provisions. No part of the Material may be used, copied,
; reproduced, modified, published, uploaded, posted, transmitted,
; distributed, or disclosed in any way without Intel's prior express
; written permission.
;
; No license under any patent, copyright, trade secret or other
; intellectual property right is granted to or conferred upon you by
; disclosure or delivery of the Materials, either expressly, by
; implication, inducement, estoppel or otherwise. Any license under
; such intellectual property rights must be express and approved by
; Intel in writing.
;
; Unless otherwise agreed by Intel in writing, you may not remove or
; alter this notice or any other notice embedded in Materials by Intel
; or Intel's suppliers or licensors in any way.

(defrule tcl-data-missing
  "Check that tcl data is available."
  (object (is-a CONFIG) (name [config]) (clck-checks $? tcl $?))
  (object (is-a NODE) (node_id ?node) (architecture x86_64)
          (role $?role&:(member$ compute ?role)|:(member$ enhanced ?role)))
  (not (object (is-a TCL) (node_id ?node)))
  =>
  (make-instance of BOOLEAN_SIGN (node_id ?node)
                 (confidence 90) (severity 90)
                 (state observed) (value TRUE)
                 (id "tcl-data-missing")
                 (remedy "tcl-data-missing-remedy")))

(defrule tcl-not-found
  "If no tcl version is found and stderr contains the string 'command
  not found', then tcl is not installed / incorrectly installed."
  (object (is-a CONFIG) (name [config]) (clck-mode certification | compliance)
          (clck-checks $? tcl $?))
  ?obj <- (object (is-a TCL) (latest-timestamp TRUE) (node_id ?hostname)
                  (stderr ?err) (version ?ver))
  (object (is-a NODE) (node_id ?hostname))
  (test (and (= (str-length ?ver) 0) (str-index "command not found" ?err)))
  =>
  (make-instance of BOOLEAN_SIGN (node_id ?hostname)
                 (confidence 90) (severity 60) (source ?obj)
                 (state observed) (value TRUE) (id "tcl-not-found")))

(defrule tcl-not-functional
  "If no tcl version is present or stderr is not empty, then tcl
   may not be functional.  If a version is present and stderr is not
   empty, use lower confidence and severity values, since the stderr
   output may be unrelated.  If no version is present and stderr is
   not empty, then tcl is definitely not functional, so use high
   confidence and severity values.  Avoid matching the 'command not
   found' case that is handled separately."
  (object (is-a CONFIG) (name [config]) (clck-mode certification | health)
          (clck-checks $? tcl $?))
  ?obj <- (object (is-a TCL) (latest-timestamp TRUE) (node_id ?hostname)
                  (stderr ?err) (version ?ver))
  (object (is-a NODE) (node_id ?hostname))
  (not (object (is-a SIGN) (id "tcl-not-found") (source ?obj)))
  (test (or (= (str-length ?ver) 0) (!= (str-length ?err) 0)))
  =>
  (if (> (str-length ?ver) 0)
      then (bind ?confidence 25)  (bind ?severity 25)
      else (bind ?confidence 90) (bind ?severity 60))
  (make-instance of BOOLEAN_SIGN (node_id ?hostname)
                 (confidence ?confidence) (severity ?severity) (source ?obj)
                 (state observed) (value TRUE) (id "tcl-not-functional")
                 (args ?err)))

(defrule tcl-not-icr
  "If the tcl version is less than 8.4.13, then tcl is not
  Intel(R) Cluster Ready 1.3.1 compliant."
  (object (is-a CONFIG) (name [config]) (clck-mode certification | compliance)
          (clck-checks $? tcl $?))
  ?obj <- (object (is-a TCL) (latest-timestamp TRUE) (node_id ?hostname)
                  (version ?ver&:(!= (str-length ?ver) 0)))
  (object (is-a NODE) (node_id ?hostname))
  (test (< (vercmp ?ver "8.4.13") 0))
  =>
  (make-instance of BOOLEAN_SIGN (node_id ?hostname)
                 (confidence 90) (severity 80) (source ?obj)
                 (state observed) (value TRUE) (id "tcl-not-icr")
                 (args (create$ ?ver "8.4.13"))))

(defrule tcl-not-uniform
  "If the tcl version is not the same as at least 90% of the other nodes,
  then the node should be flagged as non-uniform.  The fewer other
  nodes that have the same tcl version increases the confidence
  that the node with the different version is incorrect."
  (object (is-a CONFIG) (name [config]) (clck-mode certification | health)
          (clck-checks $? tcl $?))
  ?obj <- (object (is-a TCL) (latest-timestamp TRUE) (node_id ?node)
                  (version ?ver&:(!= (str-length ?ver) 0))
                  (multiset_control TRUE))
  (object (is-a NODE) (node_id ?node))
  (test (< (send ?obj fraction (send ?obj uniformity-key) ?ver) 0.9))
  =>
  (bind ?fraction (* 100 (send ?obj fraction (send ?obj uniformity-key) ?ver)))
  (make-instance of BOOLEAN_SIGN (node_id ?node)
                 (confidence (- 100 ?fraction)) (severity 40) (source ?obj)
                 (state observed) (value TRUE) (id "tcl-not-uniform")
                 (args (create$ ?fraction ?ver))))
