#!/bin/bash

HPL_DAT() {
    local tmpdir=$1
    local n=$2
    local nb=$3
    local p=$4
    local q=$5

    local HPL_DAT="${tmpdir}/HPL.dat"

    cat > $HPL_DAT <<EOF
HPLinpack benchmark input file
Innovative Computing Laboratory, University of Tennessee
HPL.out      output file name (if any)
6            device out (6=stdout,7=stderr,file)
1            # of problems sizes (N)
$n           Ns
1            # of NBs
$nb          NBs
0            PMAP process mapping (0=Row-,1=Column-major)
1            # of process grids (P x Q)
$p           Ps
$q           Qs
16.0         threshold
1            # of panel fact
2            PFACTs (0=left, 1=Crout, 2=Right)
1            # of recursive stopping criterium
4            NBMINs (>= 1)
1            # of panels in recursion
2            NDIVs
1            # of recursive panel fact.
1            RFACTs (0=left, 1=Crout, 2=Right)
1            # of broadcast
1            BCASTs (0=1rg,1=1rM,2=2rg,3=2rM,4=Lng,5=LnM)
1            # of lookahead depth
0            DEPTHs (>=0)
2            SWAP (0=bin-exch,1=long,2=mix)
64           swapping threshold
0            L1 in (0=transposed,1=no-transposed) form
0            U  in (0=transposed,1=no-transposed) form
1            Equilibration (0=no,1=yes)
8            memory alignment in double (> 0)
EOF
}

HPL_N() {
    local mem_size=$1
    local mem_percent=$2
    local nb=$3
    local nnodes=$4

    local N_sqr=$(( $mem_percent * $mem_size * 1024 * $nnodes / 800 ))
    local N_sqrt=`echo "$N_sqr $nb" | awk '{ printf("%d", sqrt($1) / $2) }'`
    N=$(( $N_sqrt * $nb ))
}

# Maximize P subject to the following constraints:
#  1. P*Q = total number of processes (all)
#  2. P <= Q
#  3. P and Q are both integers
HPL_PQ() {
    local all=$1
    local pp

    for pp in `seq 1 $all`; do
        local qq=$(( $all / $pp ))
        if [ $(( $pp * $qq )) -eq $all ] && [ $pp -le $qq ]; then
            P=$pp
            Q=$qq
        fi
    done
}

if [ ! -f "${CLCK_DP_NODEFILE}" ] ; then
    echo "Missing or invalid nodefile: '${CLCK_DP_NODEFILE}'" >&2
    exit 1
fi

. $(dirname $0)/../../functions

# setup mpi environment
mpivars

# TODO : Verify that the percentage is between 1 and 80
if [ "$CLCK_PROVIDER_HPL_CLUSTER_PERCENT_MEMORY" == "" ]; then
    mem_percent=2
else
    mem_percent=${CLCK_PROVIDER_HPL_CLUSTER_PERCENT_MEMORY}
fi

NNODES=`cat $CLCK_DP_NODEFILE | wc -l`
TMPDIR=`mktemp -d --tmpdir hpl.XXXXXX`

# get the minimum memory size per node
memory_size_cluster_min $CLCK_DP_NODEFILE $(dirname $0) 'KB'

# get the minimum number of cores per node
num_physical_cores_cluster_min $CLCK_DP_NODEFILE $(dirname $0)

# get the minimum number of coprocessors per node
mic_coprocessor_count_cluster_min $CLCK_DP_NODEFILE $(dirname $0)

# Set ppn, NB, binary location, and calculate P & Q
if [[ $min_mic_count -gt 0 ]] ; then
    #Xeon Phi
    ppn=1
    NB=256
    HPL_PQ $NNODES
    XHPL="$(dirname $0)/intel64/xhpl_offload_intel64_dynamic"
else
    #Xeon
    NB=168
    HPL_PQ $(( $NNODES * $num_physical_cores_cluster_min ))
    ppn=${num_physical_cores_cluster_min}
    XHPL="$(dirname $0)/intel64/xhpl_intel64_dynamic"
fi

# calculate N
HPL_N $memory_size_cluster_min $mem_percent $NB $NNODES

# create HPL.dat
HPL_DAT $TMPDIR $N $NB $P $Q

# Configure MPI to timeout before CLCK does to give MPI a chance to cleanup
if [[ ! -z $CLCK_TIMEOUT && $CLCK_TIMEOUT -gt 5 ]]; then
    export I_MPI_JOB_TIMEOUT=$(( $CLCK_TIMEOUT - 5 ))
fi

# Configure MPI to cleanup
export I_MPI_MPIRUN_CLEANUP=1

# Look for user configured options sent in via environment variables

if [ "$CLCK_PROVIDER_HPL_CLUSTER_FABRICS" == "" ]; then
    MY_FABRICS=""
else
    MY_FABRICS="-genv I_MPI_FABRICS=${CLCK_PROVIDER_HPL_CLUSTER_FABRICS}"
fi

if [ "$CLCK_PROVIDER_HPL_CLUSTER_DAPL_PROVIDER" == "" ]; then
    MY_DAPL=""
else
    MY_DAPL="-genv I_MPI_DAPL_PROVIDER=${CLCK_PROVIDER_HPL_CLUSTER_DAPL_PROVIDER}"
fi

if [ "$CLCK_PROVIDER_HPL_CLUSTER_OFA_ADAPTER_NAME" == "" ]; then
    MY_OFA=""
else
    MY_OFA="-genv I_MPI_OFA_ADAPTER_NAME=${CLCK_PROVIDER_HPL_CLUSTER_OFA_ADAPTER_NAME}"
fi

if [ "$CLCK_PROVIDER_HPL_CLUSTER_TCP_NETMASK" == "" ]; then
    MY_TCP=""
else
    MY_TCP="-genv I_MPI_TCP_NETMASK=${CLCK_PROVIDER_HPL_CLUSTER_TCP_NETMASK}"
fi

if [ "$CLCK_PROVIDER_HPL_CLUSTER_TMI_PROVIDER" == "" ]; then
    MY_TMI=""
else
    MY_TMI="-genv I_MPI_TMI_PROVIDER=${CLCK_PROVIDER_HPL_CLUSTER_TMI_PROVIDER}"
fi

if [ "$CLCK_PROVIDER_HPL_CLUSTER_OPTIONS" == "" ]; then
    MY_OPTIONS=""
else
    MY_OPTIONS=${CLCK_PROVIDER_HPL_CLUSTER_OPTIONS}
fi

# run HPL
# If the user options has a I_MPI_DEBUG we over ride it with 5 anyway
mpirun $MY_OPTIONS $MY_OFA $MY_DAPL $MY_TCP $MY_TMI $MY_FABRICS -genv I_MPI_DEBUG 5 -ppn $ppn -f $CLCK_DP_NODEFILE -wdir $TMPDIR $XHPL

# cleanup
if [ -d $tmpdir ] ; then
    rm -rf $tmpdir
fi
